/*****************************************************************************
*
* Emrah Akin Sisbot
* Aug 10 
*
* File : sparkOro.h
* Brief: This file contains functions for SPARK - Ontology link
*
******************************************************************************/


#ifdef ORO

#include "server/sparkHeader.h"
#include <string>
#include <map>
#include <iostream>
#include <stdlib.h>
#include <portLib.h>
#include <set>

// These are the common includes for the ontology:
// oro.h defines the ontology API
#include "liboro/oro.h"
// oro_library.h is automatically generated from the oro ontology
// and contains all the concepts defined in it.
#include "liboro/oro_library.h"
// oro_exceptions.h defines the exceptions possibily generated by liboro.
#include "liboro/oro_exceptions.h"
// yarp_connector.h is an implementation of a conenctor to the server
// based on the YARP middleware.
#include "liboro/socket_connector.h"

using namespace std;
using namespace oro;

static IConnector *connector;
static Ontology *onto; //the accessor to the ontology singleton.

typedef struct agent_stmt {
  string agentID;
  set<Statement> stmtsToUpdate;
  set<string> stmtsToClear;
  set<Statement> stmtsToAdd;
} agent_stmt;

vector<agent_stmt> agents_stmts;

int connectOro(char *host, char *port)
{
  
  //initialize the connection to the ontology server.
  string oro_server_host(host);
  string oro_server_port(port); 
  
  try {
    connector = new SocketConnector(oro_server_host, oro_server_port);
    onto = Ontology::createWithConnector(*connector);
  } catch (ConnectorException ce) {
    // Connection problem
    return -1;
  } catch (OntologyServerException ose) {
    // Server problem
    return -2;
  }
  printf("Connected to the Ontology Server!\n");
  return 1;
} 


int updateObject(SPARK_OBJECT_STATE * objectState, SPARK_OBJECT_STATE * previousState)
{
  char tempBuffer[100];
  int i, j;
  char * robotName;
  agent_stmt AS;

  robotName = objectState->agentVisibilityState[0].agentName.name;
  
  // initialization of sets
  if ( agents_stmts.empty() ) {
    for(i=0; i<objectState->agentVisibilityStateNb; i++){
      AS.agentID = string(objectState->agentVisibilityState[i].agentName.name);
      agents_stmts.push_back(AS);
    }
  }
  
  // Object has disappeared
  if(objectState->disappeared) {
    if(!previousState->disappeared) {    
      sprintf(tempBuffer,"%s isMoving ?val", objectState->name.name);
      agents_stmts[0].stmtsToClear.insert(tempBuffer);
    
      for(i=0; i<objectState->agentVisibilityStateNb; i++){
      
	sprintf(tempBuffer,"%s isVisible ?val", objectState->name.name);
	agents_stmts[i].stmtsToClear.insert(tempBuffer);
	
	sprintf(tempBuffer,"%s isLocated ?val", objectState->name.name);
	agents_stmts[i].stmtsToClear.insert(tempBuffer);
	
	sprintf(tempBuffer,"%s looksAt ?val", objectState->agentVisibilityState[i].agentName.name);
	agents_stmts[0].stmtsToClear.insert(tempBuffer);
	
	sprintf(tempBuffer,"%s pointsAt ?val", objectState->agentVisibilityState[i].agentName.name);
	agents_stmts[0].stmtsToClear.insert(tempBuffer);
      }
    
      sprintf(tempBuffer,"%s isIn ?val", objectState->name.name);
      agents_stmts[0].stmtsToClear.insert(tempBuffer);
    
      sprintf(tempBuffer,"%s isOn ?val", objectState->name.name);
      agents_stmts[0].stmtsToClear.insert(tempBuffer);
      
      sprintf(tempBuffer,"%s isNextTo ?val", objectState->name.name);
      agents_stmts[0].stmtsToClear.insert(tempBuffer);
    }

    return TRUE;
  }
  
  // Begin to put things into ontology buffer
  // onto->bufferize();
  
  // Static - Moving
  if(previousState->motion != objectState->motion) { // if the motion has changed
    if(objectState->motion == OBJ_MOVING)
      sprintf(tempBuffer,"%s isMoving true", objectState->name.name);
    else
      sprintf(tempBuffer,"%s isMoving false", objectState->name.name);
    agents_stmts[0].stmtsToUpdate.insert(Statement(tempBuffer));
  }
  
  // Agent Visibility, Focus of attention and Spatial relation
  
  for(i=0; i<objectState->agentVisibilityStateNb; i++) {
    if(previousState->agentVisibilityState[i].visibility != objectState->agentVisibilityState[i].visibility) { // if visibility state has changed
      
      if(objectState->agentVisibilityState[i].visibility == VISIBLE)
	sprintf(tempBuffer,"%s isVisible true", objectState->name.name);
      else
	sprintf(tempBuffer,"%s isVisible false", objectState->name.name);
      agents_stmts[i].stmtsToUpdate.insert(Statement(tempBuffer));
    
      if(previousState->agentVisibilityState[i].visibilityPlacement != objectState->agentVisibilityState[i].visibilityPlacement) { // if visibility placement has changed
	if(objectState->agentVisibilityState[i].visibility == VISIBLE && 
	   objectState->agentVisibilityState[i].visibilityPlacement == FOA) {
	  // Non fuctional statement - First Clear then add
	  sprintf(tempBuffer,"%s looksAt ?val",
		  objectState->agentVisibilityState[i].agentName.name);
	  agents_stmts[i].stmtsToClear.insert(tempBuffer);	      
	  
	  sprintf(tempBuffer,"%s looksAt %s",
		  objectState->agentVisibilityState[i].agentName.name,
		  objectState->name.name);
	  agents_stmts[i].stmtsToAdd.insert(Statement(tempBuffer));
	}
      }
    }
    if(previousState->agentVisibilityState[i].pointedAt != objectState->agentVisibilityState[i].pointedAt) { // if pointat has changed
      if(objectState->agentVisibilityState[i].pointedAt) {
	// Non fuctional statement - First Clear then add
	sprintf(tempBuffer,"%s pointsAt ?val",
		objectState->agentVisibilityState[i].agentName.name);
	agents_stmts[i].stmtsToClear.insert(tempBuffer);	      
	
	sprintf(tempBuffer,"%s pointsAt %s",
		objectState->agentVisibilityState[i].agentName.name,
		objectState->name.name);
	agents_stmts[i].stmtsToAdd.insert(Statement(tempBuffer));
      } 
    }
        
    // Spatial Relations
    if(previousState->agentVisibilityState[i].spatialRelation != objectState->agentVisibilityState[i].spatialRelation) {
      switch(objectState->agentVisibilityState[i].spatialRelation){
      case 1: {sprintf(tempBuffer,"%s isLocated FRONT", objectState->name.name); break;}
      case 2: {sprintf(tempBuffer,"%s isLocated LEFT", objectState->name.name); break;}
      case 3: {sprintf(tempBuffer,"%s isLocated LEFT", objectState->name.name); break;}
      case 4: {sprintf(tempBuffer,"%s isLocated LEFT", objectState->name.name); break;}
      case 5: {sprintf(tempBuffer,"%s isLocated BACK", objectState->name.name); break;}
      case 6: {sprintf(tempBuffer,"%s isLocated RIGHT", objectState->name.name); break;}
      case 7: {sprintf(tempBuffer,"%s isLocated RIGHT", objectState->name.name); break;}
      case 8: {sprintf(tempBuffer,"%s isLocated RIGHT", objectState->name.name); break;}
      case 9: {sprintf(tempBuffer,"%s isLocated FRONT", objectState->name.name); break;}
      case 10: {sprintf(tempBuffer,"%s isLocated LEFT", objectState->name.name); break;}
      case 11: {sprintf(tempBuffer,"%s isLocated LEFT", objectState->name.name); break;}
      case 12: {sprintf(tempBuffer,"%s isLocated LEFT", objectState->name.name); break;}
      case 13: {sprintf(tempBuffer,"%s isLocated BACK", objectState->name.name); break;}
      case 14: {sprintf(tempBuffer,"%s isLocated RIGHT", objectState->name.name); break;}
      case 15: {sprintf(tempBuffer,"%s isLocated RIGHT", objectState->name.name); break;}
      case 16: {sprintf(tempBuffer,"%s isLocated RIGHT", objectState->name.name); break;}
      }
      
      agents_stmts[i].stmtsToUpdate.insert(Statement(tempBuffer)); 
    }
  }
  
 // Placement Relations
  for(i=0; i<objectState->placementLength; i++) {
    for(j=0; j<objectState->agentVisibilityStateNb; j++) {
      if( (previousState->agentVisibilityState[j].visibility != objectState->agentVisibilityState[j].visibility) ||
	  (previousState->agentVisibilityState[j].visibilityPlacement != objectState->agentVisibilityState[j].visibilityPlacement) ) {
	
	if(objectState->agentVisibilityState[j].visibility == VISIBLE && 
	   objectState->agentVisibilityState[j].visibilityPlacement == FOA) {
	  //TODO: Ideally we have to test if both objects are visible for an agent.
	  
	  if(objectState->placements[i].relation == ISIN)
	    sprintf(tempBuffer,"%s isIn %s", objectState->name.name, objectState->placements[i].object.name);
	  if(objectState->placements[i].relation == ISON)
	    sprintf(tempBuffer,"%s isOn %s", objectState->name.name, objectState->placements[i].object.name);
	  if(objectState->placements[i].relation == ISNEXTTO)
	    sprintf(tempBuffer,"%s isNextTo %s", objectState->name.name, objectState->placements[i].object.name);
	  
	  agents_stmts[j].stmtsToUpdate.insert(Statement(tempBuffer));
	} 
      }
    }
  }
  
  // With this final flush, things are sent to oro-server
  // onto->flush();
  
  return TRUE;
}

int sendToORO(GEN_STRING64 address, GEN_STRING64 port)
{
  int i;
  set<string>::iterator it;
  set<Statement>::iterator itu;

  try {
    if(!SDI_F->state.oroConnectionAv) {
      if ( connectOro(address.name, port.name) == 1 ) {
	SDI_F->state.oroConnectionAv = TRUE;
      }
      else {
	return FALSE;
      }
    }
    
    cout << "TIME:" << time(0) << endl;
    for ( i = 0; i < agents_stmts.size() ; i++) {
      cout << "FOR AGENT " << agents_stmts[i].agentID << ":" << endl;
      if ( !agents_stmts[i].stmtsToClear.empty()) {
	
	cout << "\tCLEAR contains:" << endl;
	for ( it=agents_stmts[i].stmtsToClear.begin() ; it != agents_stmts[i].stmtsToClear.end(); it++ )
	  cout << "\t " << *it << endl;
	cout << endl;
	
	//-------------
	onto->clearForAgent(agents_stmts[i].agentID, agents_stmts[i].stmtsToClear);
	agents_stmts[i].stmtsToClear.clear();
	//-------------
      }
      
      if ( !agents_stmts[i].stmtsToAdd.empty()) {
	
	cout << "\tADD contains:" <<endl;
	for ( itu=agents_stmts[i].stmtsToAdd.begin() ; itu != agents_stmts[i].stmtsToAdd.end(); itu++ )
	  cout << "\t " << *itu << endl;
	cout << endl;
	
	//------------
	onto->addForAgent(agents_stmts[i].agentID, agents_stmts[i].stmtsToAdd);
	agents_stmts[i].stmtsToAdd.clear();
	//------------
      }
      
      if ( !agents_stmts[i].stmtsToUpdate.empty()) {
	cout << "\tUPDATE contains:" <<endl;
	for ( itu=agents_stmts[i].stmtsToUpdate.begin() ; itu != agents_stmts[i].stmtsToUpdate.end(); itu++ )
	  cout << "\t " << *itu << endl;
	cout << endl;
	//-----------
	onto->updateForAgent(agents_stmts[i].agentID, agents_stmts[i].stmtsToUpdate);
	agents_stmts[i].stmtsToUpdate.clear();
	//-----------
      }
    }
  } catch (ConnectorException ce) {
    // Connection problem
    printf("%s\n",ce.what());
    SDI_F->state.oroConnectionAv = FALSE;
    return FALSE;
  } catch (OntologyServerException ose) {
    // Server problem
    printf("%s\n",ose.what());
    return FALSE;
  }
  return TRUE;
}


int sendReachabilitiesToORO(SPARK_ALL_REACHABILITIES *reachabilities, GEN_STRING64 address, GEN_STRING64 port)
{
  int i;
  char tempBuffer[100];
  
  try {
    if(!SDI_F->state.oroConnectionAv) {
      if ( connectOro(address.name, port.name) == 1 ) {
	SDI_F->state.oroConnectionAv = TRUE;
      }
      else {
	SDI_F->state.oroConnectionAv = FALSE;
	return FALSE;
      }
    }
    
    for(i=0; i<reachabilities->agentReachNb; i++) {
      if(reachabilities->agentReach[i].reachability == UNREACHABLE)
	sprintf(tempBuffer,"%s isReachable false",
		reachabilities->agentReach[i].objectName.name);
      else
	sprintf(tempBuffer,"%s isReachable true", 
		reachabilities->agentReach[i].objectName.name); 
      
      onto->updateForAgent(reachabilities->agentReach[i].agentName.name, Statement(tempBuffer));
      
    }
  } catch (ConnectorException ce) {
    // Connection problem
    printf("%s\n",ce.what());
    SDI_F->state.oroConnectionAv = FALSE;
    return FALSE;
  } catch (OntologyServerException ose) {
    // Server problem
    printf("%s\n",ose.what());
    SDI_F->state.oroConnectionAv = FALSE;
    return FALSE;
  }
  
  return TRUE;
}

#endif
